﻿using System;
using System.Diagnostics;
using System.Text.Json;
using System.Text.Json.Nodes;
using System.Text.Json.Serialization;

namespace HDLNoCGen
{
    static class Program
    {

        static void uncheckMetadata (string json_path, int stage)
        {
            string json = File.ReadAllText(json_path);
            ProjectSettings projectSettings = JsonSerializer.Deserialize<ProjectSettings>(json);

            if (stage <= 3)
            {
                projectSettings.databaseMetadata.writtenToDB = false;
            }
            if (stage <= 2)
            {
                projectSettings.quartusMetadata.quartusCompiled = false;
            }
            if (stage <= 1)
            {
                projectSettings.graphVerilogMetadata.verilogGenerated = false;
            }
            if (stage == 0)
            {
                projectSettings.graphVerilogMetadata.graphSerialized = false;
            }

            json = JsonSerializer.Serialize<ProjectSettings>(projectSettings, new JsonSerializerOptions { WriteIndented = true });
            File.WriteAllText(json_path, json);
        }

        static void Main(string[] args)
        {

            bool launch_manager = false;
            string project_name = "";
            string project_location = "";
            string project_new_name = "";
            string project_action = "o";

            bool launch_graph = false;
            string topology = "";
            List<int> generators = new List<int>();
            string queue_type = "";
            string queue_position = "";
            string arbiter_type = "";
            string algorithm = "";
            int info_width = 0;
            int queue_length = 0;
            bool create_verilog = false;

            bool launch_quartus = false;
            string device_name = "5CGXFC9E7F35C8";

            bool launch_db = false;
            string db_ip = "";
            string db_username = "";
            string db_password = "";
            string db_name = "";
            int db_port = 5432;

            bool launch_settings = false;
            string quartus_path = "";

            try
            {
                for (int i = 0; i < args.Length; i++)
                {
                    switch (args[i])
                    {
                        case "--project":
                            launch_manager = true;

                            bool project_breaker = false;
                            while (!project_breaker)
                            {
                                if (i + 1 < args.Length)
                                {
                                    switch (args[++i])
                                    {
                                        case "-n":
                                        case "--name":
                                            project_name = args[++i];
                                            break;
                                        case "-l":
                                        case "--location":
                                            project_location = args[++i];
                                            break;
                                        case "-o":
                                        case "--open":
                                            project_action = "o";
                                            break;
                                        case "-c":
                                        case "--create":
                                            project_action = "c";
                                            break;
                                        case "-e":
                                        case "--erase":
                                            project_action = "e";
                                            break;
                                        case "-r":
                                        case "--rename":
                                            project_action = "r";
                                            project_new_name = args[++i];
                                            break;
                                        default:
                                            i--;
                                            project_breaker = true;
                                            break;
                                    }
                                }
                                else
                                {
                                    project_breaker = true;
                                }
                            }
                            break;
                        case "--graph":
                            launch_graph = true;

                            bool graph_breaker = false;

                            while (!graph_breaker)
                            {
                                if (i + 1 < args.Length)
                                {
                                    switch (args[++i])
                                    {
                                        case "c":
                                            topology = "c";
                                            break;
                                        case "m":
                                            topology = "m";
                                            break;
                                        case "t":
                                            topology = "t";
                                            break;
                                        case "--queue_type":
                                            queue_type = args[++i];
                                            break;
                                        case "--queue_position":
                                            queue_position = args[++i];
                                            break;
                                        case "--arbiter_type":
                                            arbiter_type = args[++i];
                                            break;
                                        case "--algorithm":
                                            algorithm = args[++i];
                                            break;
                                        case "--info_width":
                                            try
                                            {
                                                info_width = Convert.ToInt32(args[++i]);
                                            }
                                            catch (Exception e)
                                            {
                                                Console.WriteLine("--info_width must be a number");
                                                Environment.Exit(1);
                                            }
                                            break;
                                        case "--queue_length":
                                            try
                                            {
                                                queue_length = Convert.ToInt32(args[++i]);
                                            }
                                            catch (Exception e)
                                            {
                                                Console.WriteLine("--queue_lenght must be a number");
                                                Environment.Exit(1);
                                            }
                                            break;
                                        case "-v":
                                        case "--verilog":
                                            create_verilog = true;
                                            break;
                                        default:
                                            try
                                            {
                                                generators.Add(Convert.ToInt32(args[i]));
                                            }
                                            catch (Exception e)
                                            {
                                                i--;
                                                graph_breaker = true;
                                            }
                                            break;
                                    }
                                }
                                else
                                {
                                    graph_breaker = true;
                                }
                            }
                            break;
                        case "--quartus":
                            launch_quartus = true;

                            if (i + 1 < args.Length)
                            {
                                switch (args[++i])
                                {
                                    case "-d":
                                    case "--device":
                                        device_name = args[++i];
                                        break;
                                    default:
                                        i--;
                                        break;
                                }
                            }
                            break;
                        case "--database":
                            launch_db = true;

                            bool db_breaker = false;

                            while (!db_breaker)
                            {
                                if (i + 1 < args.Length)
                                {
                                    switch (args[++i])
                                    {
                                        case "-i":
                                        case "--ip":
                                            db_ip = args[++i];
                                            break;
                                        case "-u":
                                        case "--username":
                                            db_username = args[++i];
                                            break;
                                        case "--pass":
                                        case "--password":
                                            db_password = args[++i];
                                            break;
                                        case "-n":
                                        case "--name":
                                            db_name = args[++i];
                                            break;
                                        case "--port":
                                            try
                                            {
                                                db_port = Convert.ToInt32(args[++i]);
                                            }
                                            catch (FormatException e)
                                            {
                                                Console.WriteLine("--port must be a number");
                                                Environment.Exit(1);
                                            }
                                            break;
                                        default:
                                            i--;
                                            db_breaker = true;
                                            break;
                                    }
                                }
                                else
                                {
                                    db_breaker = true;
                                }
                            }
                            break;
                        case "--settings":
                            launch_settings = true;

                            bool settings_breaker = false;
                            while (!settings_breaker)
                            {
                                if (i + 1 < args.Length)
                                {
                                    switch (args[++i])
                                    {
                                        case "--qp":
                                        case "--quartus_path":
                                            quartus_path = args[++i];
                                            break;
                                        default:
                                            i--;
                                            settings_breaker = true;
                                            break;
                                    }
                                }
                                else
                                {
                                    settings_breaker = true;
                                }
                            }
                            break;
                        default:
                            Console.WriteLine($"Argument {args[i]} does not exist. Check the spelling or check graph generator numbers");
                            Environment.Exit(1);
                            break;
                    }
                }
            }
            catch (IndexOutOfRangeException e)
            {
                Console.WriteLine("No argument after an option that requires an argument");
                Environment.Exit(1);
            }

            if (launch_manager)
            {
                using (Process manager = new Process())
                {
                    string manager_location = "../../../../../Project_manager/Project_manager/bin/Debug/net8.0";
                    string manager_arguments = $"-l {project_location} -n {project_name} -{project_action} {(project_action == "r" ? project_new_name : "")}";
                    string output_data = "";

                    manager.StartInfo.FileName = $"{manager_location}/Project_manager.exe";
                    manager.StartInfo.Arguments = manager_arguments;
                    manager.StartInfo.RedirectStandardOutput = true;
                    manager.StartInfo.WorkingDirectory = manager_location;

                    manager.OutputDataReceived += new DataReceivedEventHandler((sender, e) =>
                    {
                        if (!String.IsNullOrEmpty(e.Data))
                        {
                            output_data = $"{e.Data}\n";
                        }
                    });

                    manager.Start();
                    manager.BeginOutputReadLine();
                    manager.WaitForExit();
                    manager.Kill();

                    if (manager.ExitCode != 0)
                    {
                        Console.WriteLine("Project_manager failure:");
                        Console.Write(output_data);
                        Environment.Exit(1);
                    }
                }
            }

            if (launch_graph)
            {
                uncheckMetadata($"{project_location}/{project_name}_metadata.json", 0);
                using (Process veriloger = new Process())
                {
                    string veriloger_location = "../../../../../Graph_verilog_generator/Graph_verilog_generator/bin/Debug/net8.0";
                    string veriloger_arguments = $"-l {project_location} -n {project_name} {topology} {String.Join(" ", generators)} " +
                        $"--queue_type {queue_type} " +
                        $"--queue_position {queue_position} " +
                        $"--arbiter_type {arbiter_type} " +
                        $"--algorithm {algorithm} " +
                        $"--info_width {info_width} " +
                        $"--queue_length {queue_length} " +
                        $"{(create_verilog ? "-v" : " ")}";
                    string output_data = "";

                    veriloger.StartInfo.FileName = $"{veriloger_location}/Graph_verilog_generator.exe";
                    veriloger.StartInfo.Arguments = veriloger_arguments;
                    veriloger.StartInfo.RedirectStandardOutput = true;
                    veriloger.StartInfo.WorkingDirectory = veriloger_location;

                    veriloger.OutputDataReceived += new DataReceivedEventHandler((sender, e) =>
                    {
                        if (!String.IsNullOrEmpty(e.Data))
                        {
                            output_data = $"{e.Data}\n";
                        }
                    });

                    veriloger.Start();
                    veriloger.BeginOutputReadLine();
                    veriloger.WaitForExit();
                    veriloger.Kill();

                    if (veriloger.ExitCode != 0)
                    {
                        Console.WriteLine("Graph_verilog_generator failure:");
                        Console.Write(output_data);
                        Environment.Exit(1);
                    }
                }
            }
        }
    }
}